@@ -4,7 +4,7 @@
 //! or you could specify a configuration file. The format of configuration file is defined
 //! in mod `config`.
 
-use std::time::Duration;
+use std::{process, time::Duration};
 
 use clap::{clap_app, Arg};
 use futures::future::{self, Either};
@@ -67,6 +67,7 @@ fn main() {
 
             (@arg TCP_NO_DELAY: --("tcp-no-delay") !takes_value alias("no-delay") "Set TCP_NODELAY option for socket")
             (@arg TCP_FAST_OPEN: --("tcp-fast-open") !takes_value alias("fast-open") "Enable TCP Fast Open (TFO)")
+            (@arg TCP_KEEP_ALIVE: --("tcp-keep-alive") +takes_value {validator::validate_u64} "Set TCP keep alive timeout seconds")
 
             (@arg UDP_TIMEOUT: --("udp-timeout") +takes_value {validator::validate_u64} "Timeout seconds for UDP relay")
             (@arg UDP_MAX_ASSOCIATIONS: --("udp-max-associations") +takes_value {validator::validate_u64} "Maximum associations to be kept simultaneously for UDP relay")
@@ -199,7 +200,8 @@ fn main() {
             Some(cpath) => match Config::load_from_file(cpath, ConfigType::Local) {
                 Ok(cfg) => cfg,
                 Err(err) => {
-                    panic!("loading config \"{}\", {}", cpath, err);
+                    eprintln!("loading config \"{}\", {}", cpath, err);
+                    process::exit(common::EXIT_CODE_LOAD_CONFIG_FAILURE);
                 }
             },
             None => Config::new(ConfigType::Local),
@@ -348,6 +350,14 @@ fn main() {
             config.fast_open = true;
         }
 
+        if let Some(keep_alive) = matches.value_of("TCP_KEEP_ALIVE") {
+            config.keep_alive = Some(Duration::from_secs(
+                keep_alive
+                    .parse::<u64>()
+                    .expect("`tcp-keep-alive` is expecting an integer"),
+            ));
+        }
+
         #[cfg(any(target_os = "linux", target_os = "android"))]
         if let Some(mark) = matches.value_of("OUTBOUND_FWMARK") {
             config.outbound_fwmark = Some(mark.parse::<u32>().expect("an unsigned integer for `outbound-fwmark`"));
@@ -367,7 +377,8 @@ fn main() {
             let acl = match AccessControl::load_from_file(acl_file) {
                 Ok(acl) => acl,
                 Err(err) => {
-                    panic!("loading ACL \"{}\", {}", acl_file, err);
+                    eprintln!("loading ACL \"{}\", {}", acl_file, err);
+                    process::exit(common::EXIT_CODE_LOAD_ACL_FAILURE);
                 }
             };
             config.acl = Some(acl);
@@ -465,9 +476,15 @@ fn main() {
 
         match future::select(server, abort_signal).await {
             // Server future resolved without an error. This should never happen.
-            Either::Left((Ok(..), ..)) => panic!("server exited unexpectly"),
+            Either::Left((Ok(..), ..)) => {
+                eprintln!("server exited unexpectly");
+                process::exit(common::EXIT_CODE_SERVER_EXIT_UNEXPECTLY);
+            }
             // Server future resolved with error, which are listener errors in most cases
-            Either::Left((Err(err), ..)) => panic!("aborted with {}", err),
+            Either::Left((Err(err), ..)) => {
+                eprintln!("server aborted with {}", err);
+                process::exit(common::EXIT_CODE_SERVER_ABORTED);
+            }
             // The abort signal future resolved. Means we should just exit.
             Either::Right(_) => (),
         }
